﻿// コンパイラ指令
{$ALIGN ON}
{$BOOLEVAL OFF}
{$EXTENDEDSYNTAX ON}
{$IOCHECKS ON}
{$LONGSTRINGS ON}
{$MINENUMSIZE 1}
{$OPENSTRINGS ON}
{$OVERFLOWCHECKS OFF}
{$RANGECHECKS OFF}
{$TYPEDADDRESS OFF}
{$WRITEABLECONST ON}


//Delphi 3.0J 以降 または C++Builder 3.0J 以降かチェック
{$IFNDEF VER90}
  {$IFNDEF VER93}
    {$DEFINE VER30ORHIGHER}
  {$ENDIF}
{$ENDIF}


unit NkDIBUtils;

interface

uses Windows, Classes, SysUtils, Graphics, NkDIB;

type
  ENkDIBUtilsError = class(Exception);

  TNkTransformMatrix = record
    a11, a12, a13,
    a21, a22, a23: Extended;
  end;

  // NewX = a11 * OldX + a12 * OldY + a13;
  // NewY = a21 * OldX + a22 * OldY + a23;


  TNkTriples = array[-5..5] of TNkTriple;
  PNkTriples = ^TNkTriples;
  TNkTripleMatrix = array[-5..5] of PNkTriples;

  TNkRotation =( nkRtLeft, nkRtRight, nkRtUpsideDown);
  TNkTransFormProc = function(x, y: integer; pData: Pointer): TPoint;

{$IFNDEF VER30ORHIGHER}
  TProgressStage = (psStarting, psRunning, psEnding);
{$ENDIF}
  TNkProgressMethod = procedure(Percentage: Integer; Stage: TProgressStage)
                      of object;

  TNkFilter = function(x, y: Integer; Mat: TNkTripleMatrix;
                       pData: Pointer): TNkTriple;


function NkGetScreenBits: Integer;
function NkDIBMirrorHorz(ADIB: TNkDIB): TNkDIB;
function NkDIBMirrorVert(ADIB: TNkDIB): TNkDIB;
function NkDIBRotate(ADIB: TNkDIB; rot: TNkRotation): TNkDIB;
function NkDIBZoom(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                   ZoomedPortion: TRect): TNkDIB;
function NkDIBResize(ADIB: TNkDIB; NewWidth, NewHeight: Integer): TNkDIB;
function NkDIBTransform(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                        TransFormProc: TNkTransformProc;
                        DefaultColor: TColor;
                        Progress: TNkProgressMethod;
                        pData: Pointer): TNkDIB;
function NkDIBTransformWithMatrix(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                                  AMat: TNkTransformMatrix;
                                  DefaultColor: TColor;
                                  Progress: TNkProgressMethod): TNkDIB;
function NkDIBFilter(ADIB: TNkDIB; Filter: TNkFilter;
                     Progress: TNkProgressMethod; pData: Pointer): TNkDIB;

type
  // リージョンの操作(リージョン演算：Add)の種類
  TNkRgnOperation = (nkroAnd,  // AND
                     nkroCopy, // コピー。指定されたリージョンを
                               // そのまま TNkRegion 内に取り込む
                     nkroDiff, // 差
                     nkroOr,   // OR
                     nkroXor   // XOR
                     );
  // AddPolygon のフィルモード

  // TNkRegion は TNkRegion 内のリージョンと指定されたリージョン
  // 間で演算を行い、結果を TNkRegion 内に格納します。
  // これを Add と呼ぶことにします。

  TnkRgnFillMode  = (nkfmAlternate, nkfmWinding);

//////////
// TNkRegion: リージョンを操作するクラス

  TNkRegion = class
  private
    FHandle: HRGN;  // リージョンハンドル


    //　ハンドルで指定されたリージョンを Add する。
    procedure Add(h: HRGN; op: TNkRgnOperation); overload;
  public
    constructor Create;
    destructor Destroy; override;

    // 矩形を Add する
    procedure AddRect(x1, y1, x2, y2: Integer;
                      op: TNkRgnOperation);
    // 角丸矩形を Add する
    procedure AddRoundRect(x1, y1, x2, y2: Integer;
                           Width, Height: Integer;
                           op: TNkRgnOperation);
    // 楕円を Add する
    procedure AddEllipse(x1, y1, x2, y2: Integer;
                         op: TNkRgnOperation);
    // ポリゴンを Add する
    procedure AddPolygon(pts: array of TPoint;
                         FillMode: TNkRgnFillMode;
                         op: TNkRgnOperation);
    // リージョンを Add する
    procedure Add(Rgn: TNkRegion; op: TNkRgnOperation); overload;

    // リージョンを平行移動する
    procedure Move(x, y: Integer);

    // リージョンのコピーを取り、そのハンドルを返す。
    // 得られたハンドルは TNkRegion で管理されていないハンドル。
    function  CopyHandle: HRGN;

    // リージョンをストリーム／ファイルにセーブする。
    procedure SaveToStream(Stream: TStream);
    procedure SaveToFile(FileName: string);
    // リージョンをストリーム／ファイルからロードする。
    procedure LoadFromStream(Stream: TStream);
    procedure LoadFromFile(Filename: string);

    //　ビットマップからリージョンを作る
    procedure MakeRgnFromBitmap(Bitmap: TBitmap;
                                TransparentColor: TColor);

    // リージョンのハンドル。TNkRegion で管理されているので
    // DeleteObject してはいけない。
    property Handle: HRGN read FHandle;
  end;

implementation

uses Math;


procedure RaiseError(s: string);
begin
  Raise ENkDIBUtilsError.Create(s);
end;

function NkGetScreenBits: Integer;
var h: HDC;
begin
  h := GetDC(0);
  if h = 0 then RaiseError('NkGetScreenBits: Cannot Get Screen DC');
  try
    Result := GetDeviceCaps(h, PLANES) * GetDeviceCaps(h, BITSPIXEL);
  finally
    ReleaseDC(0, h);
  end;
end;


function NkDIBMirrorHorz(ADIB: TNkDIB): TNkDIB;
var dib, DestDIB: TNkDIB;   // 変換元、変換先 DIB
    ShouldFree: Boolean;    // True なら dib を Free する。
    x, y, k: Integer;       // 座標 ＋ α
    ps, pd: PBYTE;          // ピクセルを指すポインタ ps: 変換元 pd: 変換先
    pts, ptd: ^TNkTriple;   // ピクセルを指すポインタ pts: 変換元 ptd: 変換先
    sMask, dMask: BYTE;     // ビットマスク sMask: 変換元用 dMask: 変換先用
    ColorIndex: Word;       // カラーインデックス値
begin
  ShouldFree := False;
  DestDIB := Nil;
  dib := Nil;
  try
    // 変換元が圧縮形式のDIB なら 非圧縮のDIB を作る。
    case ADIB.PixelFormat of
      nkPf4BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf4Bit;
      end;
      nkPf8BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf8Bit;
      end
      else dib := ADIB;
    end;

    // 変換先の DIB を生成
    DestDIB := TNkDIB.Create;
    try
      DestDIB.Width := dib.Width;
      DestDIB.Height := dib.Height;
      DestDIB.BitCount := dib.BitCount;
      DestDIB.Palette := CopyPalette(dib.Palette);

      case dib.PixelFormat of
        nkPf1Bit: begin     // 1bpp の場合
          for y := 0 to dib.Height-1 do begin
            // 変換元／変換先のポインタ、マスクをセット
            ps := dib.ScanLine[y];
            pd := PByte(LongInt(DestDIB.ScanLine[y]) + (dib.Width-1) div 8);
            sMask := $80;
            dMask := $80 shr ((dib.Width-1) mod 8);

            for x := 0 to dib.Width-1 do begin
              // 1 bit コピー
              if (ps^ and sMask) <> 0 then
                pd^ := pd^ or dMask
              else
                pd^ := pd^ and not dMask;

              // ポインタとマスクを更新
              if sMask <> $01 then sMask := sMask shr 1
              else begin
                Inc(ps); sMask := $80;
              end;
              if dMask <> $80 then dMask := dMask shl 1
              else begin
                dec(pd); dMask := $01;
              end;
            end;
          end;
          Result := DestDIB;
        end;
        nkPf4Bit: begin // 4bpp の場合
          for y := 0 to dib.Height-1 do begin
            // 変換元／変換先のポインタ、補助変数(X座標)をセット
            ps := dib.ScanLine[y];
            pd := PByte(LongInt(DestDIB.ScanLine[y]) + (dib.Width-1) div 2);
            k := dib.Width-1;


            for x := 0 to dib.Width-1 do begin
              // 4bit コピー
              if (x mod 2) = 0 then ColorIndex := (ps^ and $f0) shr 4
                               else ColorIndex := ps^ and $0f;

              if (k mod 2) = 0 then pd^ := (pd^ and $0f) or (ColorIndex shl 4)
                               else pd^ := (pd^ and $f0) or ColorIndex;


              // ポインタ、補助変数を更新
              if (x mod 2) = 1 then Inc(ps);
              if (k mod 2) = 0 then Dec(pd);
              Dec(k);
            end;
          end;
          Result := DestDIB;
        end;
        nkPf8Bit: begin  // 8bpp の場合
          for y := 0 to dib.Height-1 do begin
            // 変換元／変換先のポインタをセット
            ps := dib.ScanLine[y];
            pd := PByte(LongInt(DestDIB.ScanLine[y]) + dib.Width-1);
            for x := 0 to dib.Width-1 do begin
              // コピー
              pd^ := ps^;
              Inc(ps); Dec(pd);
            end;
          end;
          Result := DestDIB;
        end;
        nkPf24Bit: begin
          for y := 0 to dib.Height-1 do begin
            // 変換元／変換先のポインタをセット
            pts := dib.ScanLine[y];
            ptd := Pointer(LongInt(DestDIB.ScanLine[y]) + (dib.Width-1)*3);
            for x := 0 to dib.Width-1 do begin
              // コピー
              ptd^ := pts^;
              Inc(pts); Dec(ptd);
            end;
          end;
          Result := DestDIB;
        end;
        else
          RaiseError('NkMakeHorzMirror: Illegal DIB Type');
      end;
    except
      DestDIB.Free;
      raise;
    end;
  finally
    if (dib <> Nil) and ShouldFree then dib.Free;
  end;
end;

function NkDIBMirrorVert(ADIB: TNkDIB): TNkDIB;
var dib, DestDIB: TNkDIB;
    ShouldFree: Boolean;
    y, h, Linelength: Integer;
    ps, pd: PBYTE;
begin
  ShouldFree := False;
  DestDIB := Nil;
  dib := Nil;
  try
    case ADIB.PixelFormat of
      nkPf4BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf4Bit;
      end;
      nkPf8BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf8Bit;
      end
      else dib := ADIB;
    end;

    DestDIB := TNkDIB.Create;
    try
      DestDIB.Width := dib.Width;
      DestDIB.Height := dib.Height;
      DestDIB.BitCount := dib.BitCount;
      DestDIB.Palette := CopyPalette(dib.Palette);
      LineLength := ((dib.BitCount * dib.Width + 31) div 32) * 4;
      h := dib.Height;
      for y := 0 to h-1 do
        System.Move(dib.Scanline[y]^, DestDIB.Scanline[h-y-1]^, LineLength);
      Result := DestDIB;
    except
      DestDIB.Free;
      raise;
    end;
  finally
    if (dib <> Nil) and ShouldFree then dib.Free;
  end;
end;

function NkDIBRotate(ADIB: TNkDIB; rot: TNkRotation): TNkDIB;
type
  PPBYTE = ^PBYTE;
  PNKTriple = ^TNkTriple;
  PPNKTRIPLE = ^PNKTriple;
var dib, DestDIB: TNkDIB;
    ShouldFree: Boolean;
    ps: PBYTE;
    pdList: Pointer;
    ppd:    PPBYTE;
    pts:    PNkTriple;
    pptd:   PPNkTriple;
    pd:     PBYTE;
    ptd:    PNkTriple;

    sMask, dMask: BYTE;
    i, x, y, h, k: Integer;
    ColorIndex: Integer;
begin
  ShouldFree := False;
  DestDIB := Nil;
  dib := Nil;
  try
    case ADIB.PixelFormat of
      nkPf4BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf4Bit;
      end;
      nkPf8BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf8Bit;
      end
      else dib := ADIB;
    end;

    DestDIB := TNkDIB.Create;
    try
      if rot = nkRtUpsideDown then begin
        DestDIB.Width := dib.Width;
        DestDIB.Height := dib.Height;
      end
      else begin
        DestDIB.Height := dib.Width;
        DestDIB.Width := dib.Height;
      end;

      DestDIB.BitCount := dib.BitCount;
      DestDIB.Palette := CopyPalette(dib.Palette);

      case rot of
        nkRtLeft: begin
          case dib.PixelFormat of
            nkPf1Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    DestDIB.ScanLine[i];

                dMask := $80;
                for y := 0 to dib.Height -1 do begin
                  sMask := $80;
                  ps := dib.Scanline[y];
                  ppd := Pointer(LongInt(pdList) + SizeOf(Pointer) *
                        (DestDIB.Height -1));
                  for x := 0 to dib.Width-1 do begin
                    if (ps^ and sMask) <> 0 then
                      ppd^^ := ppd^^ or dMask
                    else
                      ppd^^ := ppd^^ and not dMask;

                    if sMask = $01 then begin sMask := $80; Inc(ps); end
                                   else sMask := sMask shr 1;
                    if dMask = $01 then Inc(ppd^);
                    Dec(ppd);
                  end;
                  if dMask = $01 then dMask := $80 else dMask := dMask shr 1;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf4Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    DestDIB.ScanLine[i];

                for y := 0 to dib.Height -1 do begin
                  ps := dib.Scanline[y];
                  ppd := Pointer(LongInt(pdList) + SizeOf(Pointer) *
                        (DestDIB.Height -1));
                  for x := 0 to dib.Width-1 do begin
                    if (x mod 2) = 0 then
                      ColorIndex := (ps^ and $f0) shr 4
                    else
                      ColorIndex := (ps^ and $0f);

                    if (y mod 2) = 0 then
                      ppd^^ := ppd^^ or (ColorIndex shl 4)
                    else
                      ppd^^ := ppd^^ or ColorIndex;

                    if (x mod 2) = 1 then Inc(ps);
                    if (y mod 2) = 1 then Inc(ppd^);
                    Dec(ppd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf8Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    DestDIB.ScanLine[i];

                for y := 0 to dib.Height -1 do begin
                  ps := dib.Scanline[y];
                  ppd := Pointer(LongInt(pdList) + SizeOf(Pointer) *
                        (DestDIB.Height -1));
                  for x := 0 to dib.Width-1 do begin
                    ppd^^ := ps^;

                    Inc(ps);
                    Inc(ppd^);
                    Dec(ppd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf24Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPNKTriple(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    DestDIB.ScanLine[i];

                for y := 0 to dib.Height -1 do begin
                  pts := dib.Scanline[y];
                  pptd := Pointer(LongInt(pdList) + SizeOf(Pointer) *
                         (DestDIB.Height -1));
                  for x := 0 to dib.Width-1 do begin
                    pptd^^ := pts^;

                    Inc(pts);
                    Inc(pptd^);
                    Dec(pptd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            else RaiseError('NkMakeRotation: Illegal DIB Type');
          end;
        end;
        nkRtRight: begin
          case dib.PixelFormat of
            nkPf1Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    Pointer(LongInt(DestDIB.ScanLine[i]) +
                            (DestDIB.Width-1) div 8);

                dMask := $80 shr ((DestDIB.Width-1) mod 8);
                for y := 0 to dib.Height -1 do begin
                  sMask := $80;
                  ps := dib.Scanline[y];
                  ppd := pdList;
                  for x := 0 to dib.Width-1 do begin
                    if (ps^ and sMask) <> 0 then
                      ppd^^ := ppd^^ or dMask
                    else
                      ppd^^ := ppd^^ and not dMask;

                    if sMask = $01 then begin sMask := $80; Inc(ps); end
                                   else sMask := sMask shr 1;
                    if dMask = $80 then Dec(ppd^);
                    Inc(ppd);
                  end;
                  if dMask = $80 then dMask := $01 else dMask := dMask shl 1;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf4Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    Pointer(LongInt(DestDIB.ScanLine[i]) +
                            (DestDIB.Width-1) div 2);

                h := DestDIB.Width;
                for y := 0 to dib.Height -1 do begin
                  ps := dib.Scanline[y];
                  ppd := pdList;
                  for x := 0 to dib.Width-1 do begin
                    if (x mod 2) = 0 then
                      ColorIndex := (ps^ and $f0) shr 4
                    else
                      ColorIndex := (ps^ and $0f);

                    if ((h - y) mod 2) = 0 then
                      ppd^^ := ppd^^ or (ColorIndex shl 4)
                    else
                      ppd^^ := ppd^^ or ColorIndex;

                    if (x mod 2) = 1 then Inc(ps);
                    if ((h-y) mod 2) = 1 then Dec(ppd^);
                    Inc(ppd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf8Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    Pointer(LongInt(DestDIB.ScanLine[i]) +
                            (DestDIB.Width-1));

                for y := 0 to dib.Height -1 do begin
                  ps := dib.Scanline[y];
                  ppd := pdList;
                  for x := 0 to dib.Width-1 do begin
                    ppd^^ := ps^;

                    Inc(ps);
                    Dec(ppd^);
                    Inc(ppd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            nkPf24Bit: begin
              GetMem(pdList, sizeof(PBYTE)*dib.Width);
              try
                for i := 0 to DestDIB.Height-1 do
                  PPBYTE(LongInt(pdList) + SizeOf(Pointer) * i)^ :=
                    Pointer(LongInt(DestDIB.ScanLine[i]) +
                            (DestDIB.Width-1) * 3);

                for y := 0 to dib.Height -1 do begin
                  pts := dib.Scanline[y];
                  pptd := pdList;
                  for x := 0 to dib.Width-1 do begin
                    pptd^^ := pts^;

                    Inc(pts);
                    Dec(pptd^);
                    Inc(pptd);
                  end;
                end;
                Result := DestDIB;
              finally
                FreeMem(pdList)
              end;
            end;
            else RaiseError('NkMakeRotation: Illegal DIB Type');
          end;
        end;
        nkRtUpsideDown: begin
          h := DestDIB.Height;
          case dib.PixelFormat of
            nkPf1Bit: begin     // 1bpp の場合
              for y := 0 to dib.Height-1 do begin
                // 変換元／変換先のポインタ、マスクをセット
                ps := dib.ScanLine[y];
                pd := PByte(LongInt(DestDIB.ScanLine[h - y - 1]) +
                           (dib.Width-1) div 8);
                sMask := $80;
                dMask := $80 shr ((dib.Width-1) mod 8);

                for x := 0 to dib.Width-1 do begin
                  // 1 bit コピー
                  if (ps^ and sMask) <> 0 then
                    pd^ := pd^ or dMask
                  else
                    pd^ := pd^ and not dMask;

                  // ポインタとマスクを更新
                  if sMask <> $01 then sMask := sMask shr 1
                  else begin
                    Inc(ps); sMask := $80;
                  end;
                  if dMask <> $80 then dMask := dMask shl 1
                  else begin
                    dec(pd); dMask := $01;
                  end;
                end;
              end;
              Result := DestDIB;
            end;
            nkPf4Bit: begin // 4bpp の場合
              for y := 0 to dib.Height-1 do begin
                // 変換元／変換先のポインタ、補助変数(X座標)をセット
                ps := dib.ScanLine[y];
                pd := PByte(LongInt(DestDIB.ScanLine[h - y - 1]) +
                            (dib.Width-1) div 2);
                k := dib.Width-1;


                for x := 0 to dib.Width-1 do begin
                  // 4bit コピー
                  if (x mod 2) = 0 then ColorIndex := (ps^ and $f0) shr 4
                                   else ColorIndex := ps^ and $0f;

                  if (k mod 2) = 0 then pd^ := (pd^ and $0f) or (ColorIndex shl 4)
                                   else pd^ := (pd^ and $f0) or ColorIndex;


                  // ポインタ、補助変数を更新
                  if (x mod 2) = 1 then Inc(ps);
                  if (k mod 2) = 0 then Dec(pd);
                  Dec(k);
                end;
              end;
              Result := DestDIB;
            end;
            nkPf8Bit: begin  // 8bpp の場合
              for y := 0 to dib.Height-1 do begin
                // 変換元／変換先のポインタをセット
                ps := dib.ScanLine[y];
                pd := PByte(LongInt(DestDIB.ScanLine[h - y - 1]) + dib.Width-1);
                for x := 0 to dib.Width-1 do begin
                  // コピー
                  pd^ := ps^;
                  Inc(ps); Dec(pd);
                end;
              end;
              Result := DestDIB;
            end;
            nkPf24Bit: begin
              for y := 0 to dib.Height-1 do begin
                // 変換元／変換先のポインタをセット
                pts := dib.ScanLine[y];
                ptd := Pointer(LongInt(DestDIB.ScanLine[h - y - 1]) +
                              (dib.Width-1)*3);
                for x := 0 to dib.Width-1 do begin
                  // コピー
                  ptd^ := pts^;
                  Inc(pts); Dec(ptd);
                end;
              end;
              Result := DestDIB;
            end;
            else
              RaiseError('NkMakeRotation: Illegal DIB Type');
          end;
        end;
      end;
    except
      DestDIB.Free;
      raise;
    end;
  finally
    if (dib <> Nil) and ShouldFree then dib.Free;
  end;
end;

function NkDIBZoom(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                   ZoomedPortion: TRect): TNkDIB;
var NewDIB, dib: TNkDIB;
    DestCanvas, SourceCanvas: TNkDIBCanvas;
    ShouldFree: Boolean;
begin
  if (NewWidth <= 0) or (NewHeight <= 0) then
    RaiseError('NkDIBResize: Illegal NewWidth or NewHeight');


  ShouldFree := False;
  dib := Nil;
  try
    case ADIB.PixelFormat of
      nkPf4BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf4Bit;
      end;
      nkPf8BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf8Bit;
      end
      else dib := ADIB;
    end;


    NewDIB := TNkDIB.Create;
    try
      NewDIB.Width := NewWidth; NewDIB.Height := Newheight;
      case ADIB.PixelFormat of
        nkPf4BitRLE: NewDIB.PixelFormat := nkPf4Bit;
        nkPf8BitRLE: NewDIB.PixelFormat := nkPf8Bit;
        else         NewDIB.PixelFormat := ADIB.PixelFormat;
      end;
      if NewDIB.BitCount < 24 then NewDIB.Palette := CopyPalette(ADIB.Palette);
      DestCanvas := TNkDIBCanvas.Create(NewDIB);
      try
        SourceCanvas := TNkDIBCanvas.Create(DIB);
        try
          DestCanvas.CopyRect(Rect(0, 0, NewWidth, Newheight),
                              SourceCanvas, ZoomedPortion);
        finally
          SourceCanvas.Free;
        end;
      finally
        DestCanvas.Free;
      end;
    except
      NewDIB.Free;
      raise;
    end;
  finally
    if (dib <> Nil) and ShouldFree then dib.Free;
  end;
  Result := NewDIB;
end;

function NkDIBResize(ADIB: TNkDIB; NewWidth, NewHeight: Integer): TNkDIB;
var NewDIB: TNkDIB;
    Canvas: TNkDIBCanvas;
begin
  Result := NkDIBZoom(ADIB, NewWidth, NewHeight,
                      Rect(0, 0, ADIB.Width, ADIB.Height));
end;

function NkDIBTransform(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                        TransFormProc: TNkTransformProc;
                        DefaultColor: TColor;
                        Progress: TNkProgressMethod;
                        pData: Pointer): TNkDIB;
var NewDIB, DIB: TNkDIB;
    ShouldFree: Boolean;
    Width, Height: Integer;
    x, y: Integer;
    SourcePoint: TPoint;
    r11, r12, r21, r22: Extended;
begin
  if (NewWidth <= 0) or (NewHeight <= 0) then
    RaiseError('NkDIBTransform: Illegal NewWidth or NewHeight');


  ShouldFree := False;
  DIB := Nil;
  try
    // 変換元が圧縮形式のDIB なら 非圧縮のDIB を作る。
    case ADIB.PixelFormat of
      nkPf4BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf4Bit;
      end;
      nkPf8BitRLE: begin
       dib := TNkDIB.Create;
       ShouldFree := True;
       dib.Assign(ADIB);
       dib.PixelFormat := nkPf8Bit;
      end
      else dib := ADIB;
    end;

    Width := dib.Width; Height := dib.Height;
    NewDIB := TNkDIB.Create;
    try
      NewDIB.Width := NewWidth; NewDIB.Height := Newheight;
      case ADIB.PixelFormat of
        nkPf4BitRLE: NewDIB.PixelFormat := nkPf4Bit;
        nkPf8BitRLE: NewDIB.PixelFormat := nkPf8Bit;
        else         NewDIB.PixelFormat := ADIB.PixelFormat;
      end;
      if NewDIB.BitCount < 24 then NewDIB.Palette := CopyPalette(ADIB.Palette);

      if Assigned(Progress) then
        Progress(0, psStarting);
      try
        for x := 0 to NewWidth-1 do
          for y := 0 to NewHeight-1 do begin
            SourcePoint := TransFormProc(x, y, pData);
            if (SourcePoint.X >= 0) and (SourcePoint.x < Width) and
               (SourcePoint.y >= 0) and (SourcePoint.y < Height) then
              NewDIB.Pixels[x, y] := DIB.Pixels[SourcePoint.x, SourcePoint.y]
            else
              NewDIB.Pixels[x, y] := DefaultColor;

            if Assigned(Progress) then
              Progress(x * 100 div NewWidth, psRunning);
          end;
      finally
        if Assigned(Progress) then
          Progress(100, psEnding);
      end;
    except
      NewDIB.Free;
      raise;
    end;
  finally
    if (dib <> Nil) and ShouldFree then dib.Free;
  end;
  Result := NewDIB;
end;


type
  TDataForTransformWithMatrix = record
    r11, r12, r21, r22: Extended;
    TrnsMat:  TNkTransformMatrix;
  end;

function TrnsProcWithMatrix(x, y: Integer; pData: Pointer): TPoint;
var p: ^TDataForTransformWithMatrix;
begin
  p := pData;
  with p^.TrnsMat, p^ do begin
    Result.x := floor((x-a13)*r11 + (y-a23)*r12+0.5);
    Result.y := floor((x-a13)*r21 + (y-a23)*r22+0.5);
  end;
end;


function NkDIBTransformWithMatrix(ADIB: TNkDIB; NewWidth, NewHeight: Integer;
                                  AMat: TNkTransformMatrix;
                                  DefaultColor: TColor;
                                  Progress: TNkProgressMethod): TNkDIB;
var data: TDataForTransformWithMatrix;
begin
  with AMat, data do begin
    if a11*a22 = a12*a21 then
      RaiseError('NkDIBTransformWithMatrix: Illegal Matrix');

    r11 :=  a22 * (a11*a22 - a12*a21);
    r12 := -a12 * (a11*a22 - a12*a21);
    r21 := -a21 * (a11*a22 - a12*a21);
    r22 :=  a11 * (a11*a22 - a12*a21);
    TrnsMat := AMat;
  end;

  Result := NkDIBTransForm(ADIB, NewWidth, NewHeight, TrnsProcWithMatrix,
                           DefaultColor, Progress, @data);
end;


function NkDIBFilter(ADIB: TNkDIB; Filter: TNkFilter;
                     Progress: TNkProgressMethod; pData: Pointer): TNkDIB;
var MatSave, Mat: TNkTripleMatrix;
    i: Integer;
    x, y: Integer;
    NewDIB: TNkDIB;
    DestLine: PNkTripleArray;
    LineSize, BufSize: Integer;
begin
  if ADIB.PixelFormat <> nkPf24Bit then
    RaiseError('NkDIBFiletr accepts only 24Bit DIB');


  BufSize := (ADIB.Width + 10) * 3;
  LineSize := ADIB.Width * 3;

  FillChar(Mat, SizeOf(Mat), 0);

  NewDIB := TNkDIB.Create;
  try
    NewDIB.Width := ADib.Width; NewDIB.Height := ADIB.Height;
    NewDIB.BitCount := 24;

    try
      for i := -5 to 5 do MatSave[i] := Nil;
      for i:= -5 to 5  do begin
        GetMem(MatSave[i], BufSize);
        FillChar(MatSave[i]^, BufSize, 0);
        if (i < 0) or (i >= ADIB.Height) then begin
        end
        else begin
          System.Move(ADIB.ScanLine[i]^, MatSave[i]^[0], LineSize);
        end;
      end;

      if Assigned(Progress) then
        Progress(0, psStarting);
      try
        for y := 0 to ADIB.Height-1 do begin
          DestLine := NewDIB.Scanline[y];
          Mat := MatSave;
          for x := 0 to ADIB.Width-1 do begin
            DestLine^[x] := Filter(x, y, Mat, pData);
            for i := -5 to 5 do Mat[i] := PNkTriples(LongInt(Mat[i])+3);
          end;

          FreeMem(MatSave[-5]);
          MatSave[-5] := Nil;

          for i := -5 to 4 do MatSave[i] := MatSave[i+1];
          MatSave[5] := Nil;

          GetMem(MatSave[5], BufSize);
          FillChar(MatSave[5]^, BufSize, 0);

          if y + 6 < ADIB.Height then begin
            System.Move(ADIB.ScanLine[y+6]^, MatSave[5]^[0], LineSize);
          end;
          if Assigned(Progress) then
            Progress(y * 100 div ADIB.Height, psRunning);
        end;
      finally
        if Assigned(Progress) then
          Progress(100, psEnding);
      end;
    finally
      for i := -5 to 5 do
        if MatSave[i] <> Nil then
          FreeMem(MatSave[i]);
    end;
  except
    NewDIB.Free;
    raise;
  end;

  Result := NewDIB;
end;


// TRgnOperation を AddRgn 用の引数に変換するテーブル
var RgnOperationCodes: array[0..4] of Integer =
      (RGN_AND, RGN_COPY, RGN_DIFF, RGN_OR, RGN_XOR);


function CreateNullRegion: HRGN;
begin
  Result := CreateRectRgn(0, 0, 0, 0);
  if Result = 0 then
    RaiseError('TNkRegion.Add: Cannot Create Region');
end;


{ TNkRegion }

//　ハンドルで指定されたリージョンを Add する。
procedure TNkRegion.Add(h: HRGN; op: TNkRgnOperation);
var temp: HRGN;
begin
  // Add 後のリージョンを受け取るリージョンを作成
  temp := CreateNullRegion;
  if op <> nkroCopy then begin
    if CombineRgn(temp, FHandle, h,
                  RgnOperationCodes[Ord(op)]) = ERROR then
      RaiseError('TNkRegion.Add: Add Failed');
  end
  else begin
    if CombineRgn(temp, h, 0,
                  RgnOperationCodes[Ord(op)]) = ERROR then
      RaiseError('TNkRegion.Add: Add Failed');
  end;
  DeleteObject(FHandle);
  FHandle := temp;
end;

//　TNkRegion で指定されたリージョンを Add する。
procedure TNkRegion.Add(Rgn: TNkRegion; op: TNkRgnOperation);
begin Add(Rgn.handle, op); end;

// 楕円を Add する。
procedure TNkRegion.AddEllipse(x1, y1, x2, y2: Integer;
                             op: TNkRgnOperation);
var temp: HRGN;
begin
  temp := CreateEllipticRgn(x1, y1, x2, y2);
  if temp = 0 then
    RaiseError('TNkRegion.AddEllipse: Cannot Create Region');
  try Add(temp, op); finally DeleteObject(temp); end;
end;

// ポリゴンをを Add する。
procedure TNkRegion.AddPolygon(pts: array of TPoint;
                             FillMode: TNkRgnFillMode;
                             op: TNkRgnOperation);
var temp: HRGN;
begin
  if FillMode = nkfmAlternate then
    temp := CreatePolygonRgn(Pts, High(pts)+1, ALTERNATE)
  else
    temp := CreatePolygonRgn(Pts, High(pts)+1, WINDING);
  if temp = 0 then
    RaiseError('TNkRegion.AddPolygon: Cannot Create Region');
  try Add(temp, op); finally DeleteObject(temp); end;
end;

// 矩形を Add する
procedure TNkRegion.AddRect(x1, y1, x2, y2: Integer;
                          op: TNkRgnOperation);
var temp: HRGN;
begin
  temp := CreateRectRgn(x1, y1, x2, y2);
  if temp = 0 then
    RaiseError('TNkRegion.AddRect: Cannot Create Region');
  try Add(temp, op); finally DeleteObject(temp); end;
end;

// 角丸矩形を Add する
procedure TNkRegion.AddRoundRect(x1, y1, x2, y2: Integer;
                               Width, Height: Integer;
                               op: TNkRgnOperation);
var temp: HRGN;
begin
  temp := CreateRoundRectRgn(x1, y1, x2, y2, Width, Height);
  if temp = 0 then
    RaiseError('TNkRegion.AddRoundRect: Cannot Create Region');
  try Add(temp, op); finally DeleteObject(temp); end;
end;

// リージョン複製し、そのハンドルを返す。
function TNkRegion.CopyHandle: HRGN;
var temp: HRGN;
begin
  temp := CreateNullRegion;
  if CombineRgn(temp, FHandle, 0, RGN_COPY) = ERROR then
    RaiseError('TNkRegion.CopyHandle: Cannot Create Region');
  Result := Temp;
end;

// コンストラクタ
constructor TNkRegion.Create;
begin
  inherited;
  FHandle := CreateNullRegion; // NULL リージョンを作成
end;

//　デストラクタ
destructor TNkRegion.Destroy;
begin
  // リージョンを破棄する
  if FHandle <> 0 then DeleteObject(FHandle);
  inherited;
end;

// ファイル(TRGNDATA 形式)からリージョンをロードする。
procedure TNkRegion.LoadFromFile(Filename: string);
var fs: TFileStream;
begin
  fs := TFileStream.Create(FileName, fmOpenRead);
  try LoadFromStream(fs); finally fs.Free; end;
end;

const xf: TXFORM = (eM11: 1; eM12: 0; eM21: 0; eM22: 1;
                    eDx: 0; eDy: 0);

// ストリーム(TRGNDATA 形式)からリージョンをロードする。
procedure TNkRegion.LoadFromStream(Stream: TStream);
var Buf: array of Byte; // ストリーム入力用バッファ
    pBuf: PRgnData;     // リージョンデータヘッダへのポインタ。
    RectCount: Integer; // リージョンの矩形数
    DataSize: Integer;  // 読み込む矩形データの量
    Rgn, temp: HRGN; // リージョンハンドル
begin
  // 読み込み用バッファ確保
  SetLength(Buf, SizeOf(TRgnDataHeader) + Sizeof(TRect) * 4000);
  // RGNDATA のヘッダを読む
  Stream.ReadBuffer(Pointer(Buf)^, SizeOf(TRgnDataHeader));
  pBuf := Pointer(Buf);
  pBuf.rdh.nRgnSize := 0;
  RectCount := pBuf.rdh.nCount; // 矩形数を求める。

  // Win95/98 は　64KB を超えるリージョンデータを読めないので
  // 矩形を 4000個づつ読み込む。(リージョンデータは矩形の配列)
  Rgn := CreateNullRegion;
  try
    while RectCount > 0  do begin
      if RectCount > 4000 then begin
        pBuf.rdh.nCount := 4000;
        DataSize := SizeOf(TRect) * 4000;
      end
      else begin
        pBuf.rdh.nCount := RectCount;
        DataSize := SizeOf(TRect) * RectCount;
      end;
      Stream.ReadBuffer(pBuf^.Buffer, DataSize);
      // 4000個以下の矩形からリージョンを作成
      temp := ExtCreateRegion(@xf,
                              Sizeof(TRgnDataHeader) + DataSize,
                              TRGNDATA(Pointer(Buf)^));
      if temp = 0 then
        RaiseError('TNkRegion.LoadFromStream: Cannot Create Region');
      // Rgn に作ったリージョンを足しこんでゆく
      try
        if CombineRgn(Rgn, Rgn, temp, RGN_OR) = ERROR then
          RaiseError('TNkRegion.LoadFromStream: Cannot Add Region');
      finally
        DeleteObject(temp);
      end;
      Dec(RectCount, 4000);
    end;
    //　全て成功。ハンドルを更新する。
    DeleteObject(FHandle);
    FHandle := Rgn;
  except
    DeleteObject(Rgn);
  end;
end;



//　ビットマップからリージョンを作る
procedure TNkRegion.MakeRgnFromBitmap(Bitmap: TBitmap;
                                    TransparentColor: TColor);
type TTriple = record B, G, R: Byte; end;
     TTripleArray = array[0..40000000] of TTriple;
     PTripleArray = ^TTripleArray;
var  pScanline: PTripleArray;
     i: Integer;
     ms: TMemoryStream;
     x, y: Integer;
     Color: TColor;
     StartX: Integer;
     RgnDataHeader: TRgnDataHeader;
     RgnRect: TRect;
begin
  if Bitmap.PixelFormat <> pf24bit then
    RaiseError('TNkRegion.makeRgnFromBitmap Failed');

  ms := TMemoryStream.Create;
  try
    with RgnDataHeader do begin
      dwSize := SizeOf(RgnDataHeader);
      iType := RDH_RECTANGLES;
      nRgnSize := 0;
      nCount := 0;
      rcBound := Rect(0, 0, Bitmap.Width, Bitmap.Height);
    end;
    ms.WriteBuffer(RgnDataHeader, SizeOf(RgnDataHeader));
    for y := 0 to Bitmap.Height-1 do begin
      pScanLine := Bitmap.Scanline[y];
      StartX := -1;
      for x := 0 to Bitmap.Width-1 do begin
        with pScanLine^[x] do Color := RGB(R, G, B);
        if Color <> TransparentColor then begin
          if StartX = -1 then StartX := x
        end
        else begin
          if StartX <> -1 then begin
            with RgnRect do begin
              Left := StartX; Right := x;
              Top := y; Bottom := y+1;
            end;
            ms.WriteBuffer(RgnRect, SizeOf(TRect));
            StartX := -1;
            Inc(RgnDataHeader.nCount);
          end;
        end;
      end;
      if StartX <> -1 then begin
        with RgnRect do begin
          Left := StartX; Right := Bitmap.Width;
          Top := y; Bottom := y+1;
        end;
        ms.WriteBuffer(RgnRect, SizeOf(TRect));
        Inc(RgnDataHeader.nCount);
      end;
    end;
    PRgnDataHeader(ms.Memory).nCount := RgnDataHeader.nCount;
    ms.Position := 0;
    LoadFromStream(ms);
  finally
    ms.Free;
  end;
end;

procedure TNkRegion.Move(x, y: Integer);
begin
  if OffsetRgn(FHandle, x, y) = ERROR then
    RaiseError('TNkRegion.Move Failed');
end;

procedure TNkRegion.SaveToFile(FileName: string);
var fs: TFileStream;
begin
  fs := TFileStream.Create(FileName, fmCreate);
  try SaveToStream(fs); finally fs.Free; end;
end;

// ストリーム(TRGNDATA 形式)へリージョンをセーブする。
procedure TNkRegion.SaveToStream(Stream: TStream);
var Buffer: array of Byte;
begin
  SetLength(Buffer, GetRegionData(FHandle, 0, Nil));
  ZeroMemory(Buffer, Length(Buffer));
  GetRegiondata(FHandle, Length(Buffer), Pointer(Buffer));
  Stream.WriteBuffer(Pointer(Buffer)^, Length(Buffer));
end;

end.

